##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Text catalog indexes
"""
import zope.container.contained
import zope.index.text
import zope.index.text.interfaces
import zope.interface
from zope.i18nmessageid import ZopeMessageFactory as _

import zope.catalog.attribute
import zope.catalog.interfaces


class ITextIndex(zope.catalog.interfaces.IAttributeIndex,
                 zope.catalog.interfaces.ICatalogIndex):
    """
    Interface-based catalog text index.

    We redefine the fields that
    :class:`zope.catalog.interfaces.IAttributeIndex`
    defines in order to change their defaults.
    """

    interface = zope.schema.Choice(
        title=_(u"Interface"),
        description=_(
            u"Objects will be adapted to this interface. "
            u"The default is"
            u" :class:`zope.index.text.interfaces.ISearchableText`"),
        vocabulary=_("Interfaces"),
        required=False,
        default=zope.index.text.interfaces.ISearchableText,
    )

    field_name = zope.schema.NativeStringLine(
        title=_(u"Field Name"),
        description=_(
            u"Name of the field to index. Defaults to ``getSearchableText``."),
        default="getSearchableText")

    field_callable = zope.schema.Bool(
        title=_(u"Field Callable"),
        description=_(
            u"If true (the default), then the field should be called to get"
            u" the value to be indexed"),
        default=True)


@zope.interface.implementer(ITextIndex)
class TextIndex(zope.catalog.attribute.AttributeIndex,
                zope.index.text.TextIndex,
                zope.container.contained.Contained):
    """Default implementation of :class:`ITextIndex`."""



#############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Index interface-defined attributes
"""
__docformat__ = 'restructuredtext'

import zope.interface

from zope.catalog.interfaces import IAttributeIndex


@zope.interface.implementer(IAttributeIndex)
class AttributeIndex(object):
    """Index interface-defined attributes

       Mixin for indexing a particular attribute of an object after
       first adapting the object to be indexed to an interface.

       The class is meant to be mixed with a base class that defines an
       ``index_doc`` method and an ``unindex_doc`` method:

         >>> class BaseIndex(object):
         ...     def __init__(self):
         ...         self.data = []
         ...     def index_doc(self, id, value):
         ...         self.data.append((id, value))
         ...     def unindex_doc(self, id):
         ...         for n, (iid, _) in enumerate(self.data):
         ...             if id == iid:
         ...                 del self.data[n]
         ...                 break

       The class does two things. The first is to get a named field
       from an object:

         >>> class Data(object):
         ...     def __init__(self, v):
         ...         self.x = v

         >>> from zope.catalog.attribute import AttributeIndex
         >>> class Index(AttributeIndex, BaseIndex):
         ...     pass

         >>> index = Index('x')
         >>> index.index_doc(11, Data(1))
         >>> index.index_doc(22, Data(2))
         >>> index.data
         [(11, 1), (22, 2)]

       If the field value is ``None``, indexing it removes it from the
       index:

         >>> index.index_doc(11, Data(None))
         >>> index.data
         [(22, 2)]

       If the field names a method (any callable object), the results
       of calling that field can be indexed:

         >>> def z(self): return self.x + 20
         >>> Data.z = z
         >>> index = Index('z', field_callable=True)
         >>> index.index_doc(11, Data(1))
         >>> index.index_doc(22, Data(2))
         >>> index.data
         [(11, 21), (22, 22)]

       Of course, if you neglect to set ``field_callable`` when you
       index a method, it's likely that most concrete index implementations
       will raise an exception, but this class will happily pass that callable
       on:

         >>> index = Index('z')
         >>> data = Data(1)
         >>> index.index_doc(11, data)
         >>> index.data
         [(11, <bound method ...>>)]


       The class can also adapt an object to an interface before getting the
       field:

         >>> from zope.interface import Interface
         >>> class I(Interface):
         ...     pass

         >>> class Data(object):
         ...     def __init__(self, v):
         ...         self.x = v
         ...     def __conform__(self, iface):
         ...         if iface is I:
         ...             return Data2(self.x)

         >>> class Data2(object):
         ...     def __init__(self, v):
         ...         self.y = v * v

         >>> index = Index('y', I)
         >>> index.index_doc(11, Data(3))
         >>> index.index_doc(22, Data(2))
         >>> index.data
         [(11, 9), (22, 4)]

       If adapting to the interface fails, the object is not indexed:

         >>> class I2(Interface): pass
         >>> I2(Data(3), None) is None
         True
         >>> index = Index('y', I2)
         >>> index.index_doc(11, Data(3))
         >>> index.data
         []

       When you define an index class, you can define a default
       interface and/or a default interface:

         >>> class Index(AttributeIndex, BaseIndex):
         ...     default_interface = I
         ...     default_field_name = 'y'

         >>> index = Index()
         >>> index.index_doc(11, Data(3))
         >>> index.index_doc(22, Data(2))
         >>> index.data
         [(11, 9), (22, 4)]

       """

    #: Subclasses can set this to a string if they want to allow
    #: construction without passing the ``field_name``.
    default_field_name = None
    #: Subclasses can set this to an interface (a callable taking
    #: the object do index and the default value to return)
    #: if they want to allow construction that doesn't provide an
    #: ``interface``.
    default_interface = None

    def __init__(self, field_name=None, interface=None, field_callable=False,
                 *args, **kwargs):
        super(AttributeIndex, self).__init__(*args, **kwargs)
        if field_name is None and self.default_field_name is None:
            raise ValueError("Must pass a field_name")
        if field_name is None:
            self.field_name = self.default_field_name
        else:
            self.field_name = field_name
        if interface is None:
            self.interface = self.default_interface
        else:
            self.interface = interface
        self.field_callable = field_callable

    def index_doc(self, docid, object):
        """
        Derives the value to index for *object*.

        Uses the interface passed to the constructor to adapt the
        object, and then gets the field (calling it if
        ``field_callable`` was set). If the value thus found is ``None``,
        calls ``unindex_doc``. Otherwise, passes the *docid* and the value to
        the superclass's implementation of ``index_doc``.
        """
        if self.interface is not None:
            object = self.interface(object, None)
            if object is None:
                return None

        value = getattr(object, self.field_name, None)

        if value is not None and self.field_callable:
            # do not eat the exception raised below
            value = value()

        if value is None:
            # unindex the previous value!
            super(AttributeIndex, self).unindex_doc(docid)
            return None

        return super(AttributeIndex, self).index_doc(docid, value)



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests for catalog

Note that indexes &c already have test suites, we only have to check that
a catalog passes on events that it receives.
"""
import doctest
import re
import unittest

from BTrees.IFBTree import IFSet
from zope.component import eventtesting
from zope.component import provideAdapter
from zope.component import provideUtility
from zope.component import testing
from zope.component.hooks import resetHooks
from zope.component.hooks import setHooks
from zope.component.hooks import setSite
from zope.component.interfaces import ISite
from zope.container.interfaces import ISimpleReadContainer
from zope.container.traversal import ContainerTraversable
from zope.index.interfaces import IIndexSearch
from zope.index.interfaces import IIndexSort
from zope.index.interfaces import IInjection
from zope.interface import Interface
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.interface.interfaces import IComponentLookup
from zope.interface.verify import verifyObject
from zope.intid.interfaces import IIntIds
from zope.location.location import Location
from zope.site.folder import Folder
from zope.site.folder import rootFolder
from zope.site.site import LocalSiteManager
from zope.site.site import SiteManagerAdapter
from zope.testing import cleanup
from zope.testing import renormalizing
from zope.traversing import api
from zope.traversing.interfaces import ITraversable
from zope.traversing.testing import setUp as traversingSetUp

from zope.catalog.catalog import Catalog
from zope.catalog.field import FieldIndex
from zope.catalog.interfaces import ICatalog
from zope.catalog.interfaces import INoAutoIndex
from zope.catalog.interfaces import INoAutoReindex


checker = renormalizing.RENormalizing([
    # Python 3 unicode removed the "u".
    (re.compile("u('.*?')"),
     r"\1"),
    (re.compile('u(".*?")'),
     r"\1"),
])


@implementer(IIntIds)
class IntIdsStub(object):
    """A stub for IntIds."""

    def __init__(self):
        self.ids = {}
        self.objs = {}
        self.lastid = 0

    def _generateId(self):
        self.lastid += 1
        return self.lastid

    def register(self, ob):
        if ob not in self.ids:
            uid = self._generateId()
            self.ids[ob] = uid
            self.objs[uid] = ob
            return uid
        raise NotImplementedError()

    def unregister(self, ob):
        raise NotImplementedError()

    def getObject(self, uid):
        return self.objs[uid]

    def getId(self, ob):
        return self.ids[ob]

    def queryId(self, ob, default=None):
        return self.ids.get(ob, default)

    def __iter__(self):
        return iter(self.objs.keys())


@implementer(IIndexSearch, IInjection)
class StubIndex(object):
    """A stub for Index."""

    def __init__(self, field_name, interface=None):
        self._field_name = field_name
        self.interface = interface
        self.doc = {}

    def index_doc(self, docid, obj):
        self.doc[docid] = obj

    def unindex_doc(self, docid):
        del self.doc[docid]

    def apply(self, term):
        results = []
        for docid in self.doc:
            obj = self.doc[docid]
            fieldname = getattr(obj, self._field_name, '')
            if fieldname == term:
                results.append(docid)
        return IFSet(results)

    def clear(self):
        self.doc = {}


class stoopid:
    def __init__(self, **kw):
        self.__dict__ = kw


class PlacelessSetup(testing.PlacelessSetup,
                     eventtesting.PlacelessSetup):

    def setUp(self, doctesttest=None):
        testing.PlacelessSetup.setUp(self)
        eventtesting.PlacelessSetup.setUp(self)


class Test(PlacelessSetup, unittest.TestCase):

    def test_family(self):
        catalog = Catalog()
        self.assertEqual(catalog.family, Catalog.family)
        catalog = Catalog(family=self)
        self.assertEqual(catalog.family, self)

    def test_catalog_add_del_indexes(self):
        catalog = Catalog()
        verifyObject(ICatalog, catalog)
        index = StubIndex('author', None)
        catalog['author'] = index
        self.assertEqual(list(catalog.keys()), ['author'])
        index = StubIndex('title', None)
        catalog['title'] = index
        indexes = sorted(catalog.keys())
        self.assertEqual(indexes, ['author', 'title'])
        del catalog['author']
        self.assertEqual(list(catalog.keys()), ['title'])

    def _frob_intidutil(self, ints=True):
        uidutil = IntIdsStub()
        provideUtility(uidutil, IIntIds)
        # whack some objects in our little objecthub
        if ints:
            for i in range(10):
                uidutil.register("<object %s>" % i)
        uidutil.register(stoopid(simiantype='monkey', name='bobo'))
        uidutil.register(stoopid(simiantype='monkey', name='bubbles'))
        uidutil.register(stoopid(simiantype='monkey', name='ginger'))
        uidutil.register(stoopid(simiantype='bonobo', name='ziczac'))
        uidutil.register(stoopid(simiantype='bonobo', name='bobo'))
        uidutil.register(stoopid(simiantype='punyhuman', name='anthony'))
        uidutil.register(stoopid(simiantype='punyhuman', name='andy'))
        uidutil.register(stoopid(simiantype='punyhuman', name='kev'))

    def test_updateindexes(self):
        """Test a full refresh."""
        self._frob_intidutil()
        catalog = Catalog()
        catalog['author'] = StubIndex('author', None)
        catalog['title'] = StubIndex('author', None)
        catalog.updateIndexes()
        for index in catalog.values():
            checkNotifies = index.doc
            self.assertEqual(len(checkNotifies), 18)

    def test_updateindex(self):
        """Test a full refresh."""
        self._frob_intidutil()
        catalog = Catalog()
        catalog['author'] = StubIndex('author', None)
        catalog['title'] = StubIndex('author', None)
        catalog.updateIndex(catalog['author'])
        checkNotifies = catalog['author'].doc
        self.assertEqual(len(checkNotifies), 18)
        checkNotifies = catalog['title'].doc
        self.assertEqual(len(checkNotifies), 0)

    def test_basicsearch(self):
        """Test the simple search results interface."""
        self._frob_intidutil(ints=0)
        catalog = Catalog()
        catalog['simiantype'] = StubIndex('simiantype', None)
        idx = catalog['name'] = StubIndex('name', None)
        catalog.updateIndexes()

        res = catalog.searchResults(simiantype='monkey')
        names = sorted([x.name for x in res])
        self.assertEqual(len(names), 3)
        self.assertEqual(names, ['bobo', 'bubbles', 'ginger'])

        res = catalog.searchResults(name='bobo')
        names = [x.simiantype for x in res]
        names.sort()
        self.assertEqual(len(names), 2)
        self.assertEqual(names, ['bonobo', 'monkey'])

        res = catalog.searchResults(name='bobo', _reverse=True)
        names = [x.simiantype for x in res]
        self.assertEqual(len(names), 2)
        self.assertEqual(names, ['bonobo', 'monkey'])

        res = catalog.searchResults(name='bobo', _limit=1)
        names = [x.simiantype for x in res]
        self.assertEqual(len(names), 1)
        self.assertEqual(names, ['monkey'])

        with self.assertRaises(ValueError):
            catalog.searchResults(name='bobo', _sort_index='name')

        alsoProvides(idx, IIndexSort)

        def sort(results, limit=0, reverse=0):
            return sorted(results, reverse=reverse)
        idx.sort = sort

        res = catalog.searchResults(
            name='bobo', _sort_index='name', _reverse=False)
        names = [x.simiantype for x in res]
        self.assertEqual(len(names), 2)
        self.assertEqual(names, ['monkey', 'bonobo'])

        res = catalog.searchResults(name='doc')
        self.assertEqual(len(res), 0)

        res = catalog.searchResults(simiantype='punyhuman', name='anthony')
        self.assertEqual(len(res), 1)
        ob = next(iter(res))
        self.assertEqual((ob.name, ob.simiantype), ('anthony', 'punyhuman'))

        res = catalog.searchResults(simiantype='ape', name='bobo')
        self.assertEqual(len(res), 0)

        res = catalog.searchResults(simiantype='ape', name='mwumi')
        self.assertEqual(len(res), 0)
        self.assertRaises(KeyError, catalog.searchResults,
                          simiantype='monkey', hat='beret')

        catalog.clear()
        res = catalog.searchResults(name='bobo')
        self.assertEqual(len(res), 0)

        class BadIndex(object):
            def apply(self, _q):
                return None
        catalog['stub'] = BadIndex()
        res = catalog.searchResults(stub='foo')
        self.assertIsNone(res)


@implementer(ICatalog)
class CatalogStub:

    def __init__(self):
        self.regs = []
        self.unregs = []

    def index_doc(self, docid, doc):
        self.regs.append((docid, doc))

    def unindex_doc(self, docid):
        self.unregs.append(docid)


class Stub(Location):
    pass


class TestEventSubscribers(unittest.TestCase):

    def setUp(self):
        self.root = placefulSetUp(True)
        sm = self.root.getSiteManager()
        self.utility = addUtility(sm, '', IIntIds, IntIdsStub())
        self.cat = addUtility(sm, '', ICatalog, CatalogStub())
        setSite(self.root)

    def tearDown(self):
        placefulTearDown()

    def test_indexDocSubscriber(self):
        from zope.container.contained import ObjectAddedEvent
        from zope.intid.interfaces import IntIdAddedEvent

        from zope.catalog.catalog import indexDocSubscriber

        ob = Stub()
        ob2 = Stub()
        ob3 = Stub()
        alsoProvides(ob3, INoAutoIndex)

        self.root['ob'] = ob
        self.root['ob2'] = ob2
        self.root['ob3'] = ob2

        id = self.utility.register(ob)
        self.utility.register(ob3)
        indexDocSubscriber(IntIdAddedEvent(ob, ObjectAddedEvent(ob2)))
        indexDocSubscriber(IntIdAddedEvent(ob3, ObjectAddedEvent(ob3)))

        self.assertEqual(self.cat.regs, [(id, ob)])
        self.assertEqual(self.cat.unregs, [])

    def test_reindexDocSubscriber(self):
        from zope.lifecycleevent import ObjectModifiedEvent

        from zope.catalog.catalog import reindexDocSubscriber

        ob = Stub()

        self.root['ob'] = ob
        self.root['ob'] = ob

        self.utility.register(ob)

        reindexDocSubscriber(ObjectModifiedEvent(ob))

        self.assertEqual(self.cat.regs, [(1, ob)])
        self.assertEqual(self.cat.unregs, [])

        ob2 = Stub()
        ob3 = Stub()
        alsoProvides(ob3, INoAutoReindex)

        self.root['ob3'] = ob3

        reindexDocSubscriber(ObjectModifiedEvent(ob2))
        reindexDocSubscriber(ObjectModifiedEvent(ob3))
        self.assertEqual(self.cat.regs, [(1, ob)])
        self.assertEqual(self.cat.unregs, [])

    def test_unindexDocSubscriber(self):
        from zope.container.contained import ObjectRemovedEvent
        from zope.intid.interfaces import IntIdRemovedEvent

        from zope.catalog.catalog import unindexDocSubscriber

        ob = Stub()
        ob2 = Stub()
        ob3 = Stub()
        self.root['ob'] = ob
        self.root['ob2'] = ob2
        self.root['ob3'] = ob3

        id = self.utility.register(ob)

        unindexDocSubscriber(
            IntIdRemovedEvent(ob2, ObjectRemovedEvent(ob3)))
        self.assertEqual(self.cat.unregs, [])
        self.assertEqual(self.cat.regs, [])

        unindexDocSubscriber(
            IntIdRemovedEvent(ob, ObjectRemovedEvent(ob3)))
        self.assertEqual(self.cat.unregs, [id])
        self.assertEqual(self.cat.regs, [])


class TestIndexUpdating(unittest.TestCase):
    """Issue #466: When reindexing a catalog it takes all objects from
    the nearest IntId utility. This is a problem when IntId utility
    lives in another site than the one.

    To solve this issue we simply check whether the objects are living
    within the nearest site.
    """

    def setUp(self):
        placefulSetUp(True)

        from zope.catalog.catalog import Catalog

        self.root = buildSampleFolderTree()

        subfolder = self.root[u'folder1'][u'folder1_1']
        root_sm = self.root_sm = createSiteManager(self.root)
        local_sm = self.local_sm = createSiteManager(subfolder)
        self.utility = addUtility(root_sm, '', IIntIds, IntIdsStub())
        self.cat = addUtility(local_sm, '', ICatalog, Catalog())
        self.cat['name'] = StubIndex('__name__', None)

        for obj in self.iterAll(self.root):
            self.utility.register(obj)

    def tearDown(self):
        placefulTearDown()

    def iterAll(self, container):
        from zope.container.interfaces import IContainer
        for value in container.values():
            yield value
            if IContainer.providedBy(value):
                for obj in self.iterAll(value):
                    yield obj

    def test_visitSublocations(self):
        """ Test the iterContained method which should return only the
        sublocations which are registered by the IntIds.
        """

        names = sorted([ob.__name__ for i,
                        ob in self.cat._visitSublocations()])
        self.assertEqual(names, [u'folder1_1', u'folder1_1_1', u'folder1_1_2'])

    def test_updateIndex(self):
        """ Setup a catalog deeper within the containment hierarchy
        and call the updateIndexes method. The indexed objects should should
        be restricted to the sublocations.
        """
        self.cat.updateIndexes()
        index = self.cat['name']
        names = sorted([ob.__name__ for i, ob in index.doc.items()])
        self.assertEqual(names, [u'folder1_1', u'folder1_1_1', u'folder1_1_2'])

    def test_optimizedUpdateIndex(self):
        """ Setup a catalog deeper within the containment hierarchy together
        with its intid utility. The catalog will not visit sublocations
        because the intid utility can not contain objects outside the site
        where it is registered.
        """
        utility = addUtility(self.local_sm, '', IIntIds, IntIdsStub())
        subfolder = self.root[u'folder1'][u'folder1_1']
        for obj in self.iterAll(subfolder):
            utility.register(obj)

        self.cat.updateIndexes()
        index = self.cat['name']
        names = sorted([ob.__name__ for i, ob in index.doc.items()])
        self.assertEqual(names, [u'folder1_1_1', u'folder1_1_2'])

    def test_index_added(self):
        from zope.catalog.catalog import indexAdded
        index = self.cat['name']

        indexAdded(index, None)

        names = sorted([ob.__name__ for i, ob in index.doc.items()])
        self.assertEqual(names, [u'folder1_1', u'folder1_1_1', u'folder1_1_2'])


class TestSubSiteCatalog(unittest.TestCase):
    """If a catalog is defined in a sub site and the hooks.setSite was
    not set the catalog will not be found unless the context in
    getAllUtilitiesRegisteredFor is set.
    """

    def setUp(self):
        placefulSetUp(True)

        from zope.catalog.catalog import Catalog

        self.root = buildSampleFolderTree()

        self.subfolder = self.root[u'folder1'][u'folder1_1']
        root_sm = self.root_sm = createSiteManager(self.root)
        local_sm = self.local_sm = createSiteManager(self.subfolder)
        self.utility = addUtility(root_sm, '', IIntIds, IntIdsStub())
        self.cat = addUtility(local_sm, '', ICatalog, Catalog())
        self.cat['name'] = StubIndex('__name__', None)

        for obj in self.iterAll(self.root):
            self.utility.register(obj)

    def tearDown(self):
        placefulTearDown()

    def iterAll(self, container):
        from zope.container.interfaces import IContainer
        for value in container.values():
            yield value
            if IContainer.providedBy(value):
                for obj in self.iterAll(value):
                    yield obj

    def test_Index(self):
        """ Setup a catalog deeper within the containment hierarchy
        and call the updateIndexes method. The indexed objects should should
        be restricted to the sublocations.
        """
        from zope.container.contained import ObjectAddedEvent

        from zope.catalog.catalog import indexDocSubscriber

        ob = Stub()
        self.subfolder['ob'] = ob

        self.utility.register(ob)

        setSite(self.subfolder)
        res = self.cat.searchResults(name='ob')
        self.assertEqual(len(res), 0)

        setSite(None)
        indexDocSubscriber(ObjectAddedEvent(ob))

        setSite(self.subfolder)
        res = self.cat.searchResults(name='ob')
        self.assertEqual(len(res), 1)

    def test_updateIndex(self):
        """ Setup a catalog deeper within the containment hierarchy
        and call the updateIndexes method. The indexed objects should should
        be restricted to the sublocations.
        """
        from zope.lifecycleevent import ObjectModifiedEvent

        from zope.catalog.catalog import reindexDocSubscriber

        ob = Stub()
        self.subfolder['ob'] = ob

        self.utility.register(ob)

        setSite(self.subfolder)
        res = self.cat.searchResults(name='ob')
        self.assertEqual(len(res), 0)

        setSite(None)
        reindexDocSubscriber(ObjectModifiedEvent(ob))

        setSite(self.subfolder)
        res = self.cat.searchResults(name='ob')
        self.assertEqual(len(res), 1)

    def test_UnIndex(self):
        """ Setup a catalog deeper within the containment hierarchy
        and call the updateIndexes method. The indexed objects should should
        be restricted to the sublocations.
        """
        from zope.container.contained import ObjectAddedEvent
        from zope.container.contained import ObjectRemovedEvent

        from zope.catalog.catalog import indexDocSubscriber
        from zope.catalog.catalog import unindexDocSubscriber

        ob = Stub()
        self.subfolder['ob'] = ob

        self.utility.register(ob)

        setSite(self.subfolder)
        res = self.cat.searchResults(name='ob')
        self.assertEqual(len(res), 0)

        setSite(None)
        indexDocSubscriber(ObjectAddedEvent(ob))

        setSite(self.subfolder)
        res = self.cat.searchResults(name='ob')
        self.assertEqual(len(res), 1)

        setSite(None)
        unindexDocSubscriber(ObjectRemovedEvent(ob))

        setSite(self.subfolder)
        res = self.cat.searchResults(name='ob')
        self.assertEqual(len(res), 0)


class TestCatalogBugs(PlacelessSetup, unittest.TestCase):
    """I found that z.a.catalog, AttributeIndex failed to remove the previous
    value/object from the index IF the NEW value is None.
    """

    def test_updateIndexWithNone(self):
        uidutil = IntIdsStub()
        provideUtility(uidutil, IIntIds)

        catalog = Catalog()
        index = FieldIndex('author', None)
        catalog['author'] = index

        ob1 = stoopid(author="joe")
        ob1id = uidutil.register(ob1)
        catalog.index_doc(ob1id, ob1)

        res = catalog.searchResults(author=('joe', 'joe'))
        names = sorted([x.author for x in res])
        self.assertEqual(len(names), 1)
        self.assertEqual(names, ['joe'])

        ob1.author = None
        catalog.index_doc(ob1id, ob1)

        # the index must be empty now because None values are never indexed
        res = catalog.searchResults(author=(None, None))
        self.assertEqual(len(res), 0)

    def test_updateIndexFromCallableWithNone(self):
        uidutil = IntIdsStub()
        provideUtility(uidutil, IIntIds)

        catalog = Catalog()
        index = FieldIndex('getAuthor', None, field_callable=True)
        catalog['author'] = index

        ob1 = stoopidCallable(author="joe")

        ob1id = uidutil.register(ob1)
        catalog.index_doc(ob1id, ob1)

        res = catalog.searchResults(author=('joe', 'joe'))
        names = sorted([x.author for x in res])
        self.assertEqual(len(names), 1)
        self.assertEqual(names, ['joe'])

        ob1.author = None
        catalog.index_doc(ob1id, ob1)

        # the index must be empty now because None values are never indexed
        res = catalog.searchResults(author=(None, None))
        self.assertEqual(len(res), 0)


class stoopidCallable(object):
    def __init__(self, **kw):
        # leave the door open to not to set self.author
        self.__dict__.update(kw)

    def getAuthor(self):
        return self.author


class TestIndexRaisingValueGetter(PlacelessSetup, unittest.TestCase):
    """ """

    def test_IndexRaisingValueGetter(self):
        """We can have indexes whose values are determined by callable
        methods.
        Raising an exception in the method should not be silently ignored
        That would cause index corruption -- the index would be out of sync"""
        uidutil = IntIdsStub()
        provideUtility(uidutil, IIntIds)

        catalog = Catalog()
        index = FieldIndex('getAuthor', None, field_callable=True)
        catalog['author'] = index

        ob1 = stoopidCallable(author="joe")
        ob1id = uidutil.register(ob1)
        catalog.index_doc(ob1id, ob1)

        res = catalog.searchResults(author=('joe', 'joe'))
        names = sorted([x.author for x in res])
        self.assertEqual(len(names), 1)
        self.assertEqual(names, ['joe'])

        ob2 = stoopidCallable()  # no author here, will raise AttributeError
        ob2id = uidutil.register(ob2)
        with self.assertRaises(AttributeError):
            catalog.index_doc(ob2id, ob2)


class TestAttributeIndex(cleanup.CleanUp,
                         unittest.TestCase):

    def setUp(self):
        super(TestAttributeIndex, self).setUp()
        from zope.schema import interfaces
        from zope.schema import vocabulary

        from zope import interface

        @interface.implementer(interfaces.ISource)
        class SimpleVocabulary(object):
            def __init__(self, context):
                self.context = context

        registry = vocabulary.getVocabularyRegistry()
        registry.register('Interfaces', SimpleVocabulary)

    def test_invalid_constructor(self):
        from zope.catalog.attribute import AttributeIndex
        with self.assertRaises(ValueError):
            AttributeIndex()

    def test_index_doc_interface_returns_none(self):
        from zope.catalog.attribute import AttributeIndex

        def interface(_s, _o):
            return None
        idx = AttributeIndex(field_name='foo', interface=interface)

        self.assertIsNone(idx.index_doc(1, self))

    def test_field_name(self):
        from zope.schema import getValidationErrors

        from zope.catalog.attribute import AttributeIndex
        from zope.catalog.interfaces import IAttributeIndex

        # native string
        idx = AttributeIndex(field_name='foo')
        verifyObject(IAttributeIndex, idx)
        good_name = b'foo' if bytes is str else u'foo'
        bad_name = u'foo' if bytes is str else b'foo'

        idx = AttributeIndex(field_name=good_name)
        verifyObject(IAttributeIndex, idx)
        self.assertEqual(getValidationErrors(IAttributeIndex, idx),
                         [])

        idx = AttributeIndex(field_name=bad_name)
        verifyObject(IAttributeIndex, idx)
        self.assertEqual(getValidationErrors(IAttributeIndex, idx)[0][0],
                         'field_name')


class TestTextIndex(unittest.TestCase):

    def test_constructor(self):
        from zope.catalog.text import ITextIndex
        from zope.catalog.text import TextIndex

        verifyObject(ITextIndex, TextIndex(field_name='foo'))


class TestKeywordIndex(unittest.TestCase):

    def test_constructor(self):
        from zope.catalog.keyword import CaseInsensitiveKeywordIndex
        from zope.catalog.keyword import IKeywordIndex
        from zope.catalog.keyword import KeywordIndex

        verifyObject(IKeywordIndex, KeywordIndex(field_name='foo'))
        verifyObject(
            IKeywordIndex,
            CaseInsensitiveKeywordIndex(
                field_name='foo'))


# ------------------------------------------------------------------------
# placeful setUp/tearDown
def placefulSetUp(site=False):
    testing.setUp()
    eventtesting.setUp()
    traversingSetUp()
    setHooks()
    provideAdapter(ContainerTraversable,
                   (ISimpleReadContainer,), ITraversable)
    provideAdapter(SiteManagerAdapter, (Interface,), IComponentLookup)

    if site:
        root = rootFolder()
        createSiteManager(root, setsite=True)
        return root


def placefulTearDown():
    resetHooks()
    setSite()
    testing.tearDown()


# ------------------------------------------------------------------------
# setup site manager
def createSiteManager(folder, setsite=False):
    if not ISite.providedBy(folder):
        folder.setSiteManager(LocalSiteManager(folder))
    if setsite:
        setSite(folder)
    return api.traverse(folder, "++etc++site")


# ------------------------------------------------------------------------
# Local Utility Addition
def addUtility(sitemanager, name, iface, utility, suffix=''):
    """Add a utility to a site manager

    This helper function is useful for tests that need to set up utilities.
    """
    folder_name = (name or (iface.__name__ + 'Utility')) + suffix
    default = sitemanager['default']
    default[folder_name] = utility
    utility = default[folder_name]
    sitemanager.registerUtility(utility, iface, name)
    return utility


# ------------------------------------------------------------------------
# Sample Folder Creation
def buildSampleFolderTree():
    # set up a reasonably complex folder structure
    #
    #     ____________ rootFolder ______________________________
    #    /                                    \                 \
    # folder1 __________________            folder2           folder3
    #   |                       \             |                 |
    # folder1_1 ____           folder1_2    folder2_1         folder3_1
    #   |           \            |            |
    # folder1_1_1 folder1_1_2  folder1_2_1  folder2_1_1

    root = rootFolder()
    root[u'folder1'] = Folder()
    root[u'folder1'][u'folder1_1'] = Folder()
    root[u'folder1'][u'folder1_1'][u'folder1_1_1'] = Folder()
    root[u'folder1'][u'folder1_1'][u'folder1_1_2'] = Folder()
    root[u'folder1'][u'folder1_2'] = Folder()
    root[u'folder1'][u'folder1_2'][u'folder1_2_1'] = Folder()
    root[u'folder2'] = Folder()
    root[u'folder2'][u'folder2_1'] = Folder()
    root[u'folder2'][u'folder2_1'][u'folder2_1_1'] = Folder()
    root[u"\N{CYRILLIC SMALL LETTER PE}"
         u"\N{CYRILLIC SMALL LETTER A}"
         u"\N{CYRILLIC SMALL LETTER PE}"
         u"\N{CYRILLIC SMALL LETTER KA}"
         u"\N{CYRILLIC SMALL LETTER A}3"] = Folder()
    root[u"\N{CYRILLIC SMALL LETTER PE}"
         u"\N{CYRILLIC SMALL LETTER A}"
         u"\N{CYRILLIC SMALL LETTER PE}"
         u"\N{CYRILLIC SMALL LETTER KA}"
         u"\N{CYRILLIC SMALL LETTER A}3"][
        u"\N{CYRILLIC SMALL LETTER PE}"
        u"\N{CYRILLIC SMALL LETTER A}"
        u"\N{CYRILLIC SMALL LETTER PE}"
        u"\N{CYRILLIC SMALL LETTER KA}"
        u"\N{CYRILLIC SMALL LETTER A}3_1"] = Folder()

    return root


def test_suite():
    suite = unittest.TestSuite((
        unittest.defaultTestLoader.loadTestsFromName(__name__),
        doctest.DocTestSuite('zope.catalog.attribute',
                             optionflags=doctest.ELLIPSIS),
    ))

    return suite



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Keyword catalog index
"""
import zope.container.contained
import zope.index.keyword
import zope.interface

import zope.catalog.attribute
import zope.catalog.interfaces


class IKeywordIndex(zope.catalog.interfaces.IAttributeIndex,
                    zope.catalog.interfaces.ICatalogIndex):
    """Interface-based catalog keyword index"""


@zope.interface.implementer(IKeywordIndex)
class KeywordIndex(zope.catalog.attribute.AttributeIndex,
                   zope.index.keyword.KeywordIndex,
                   zope.container.contained.Contained):
    """
    Default implementation of :class:`IKeywordIndex`.
    """


@zope.interface.implementer(IKeywordIndex)
class CaseInsensitiveKeywordIndex(
        zope.catalog.attribute.AttributeIndex,
        zope.index.keyword.CaseInsensitiveKeywordIndex,
        zope.container.contained.Contained):
    """
    A kind of :class:`IKeywordIndex` that is not sensitive to case.
    """



##############################################################################
#
# Copyright (c) 2006 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
# This package is developed by the Zope Toolkit project, documented here:
# http://docs.zope.org/zopetoolkit
# When developing and releasing this package, please follow the documented
# Zope Toolkit policies as described by this documentation.
##############################################################################
"""Setup for zope.catalog package
"""
import os

from setuptools import find_packages
from setuptools import setup


def read(*rnames):
    with open(os.path.join(os.path.dirname(__file__), *rnames)) as f:
        return f.read()


tests_require = [
    'zope.site',
    'zope.testing',
    'zope.testrunner',
]

setup(name='zope.catalog',
      version='4.4.1',
      author='Zope Corporation and Contributors',
      author_email='zope-dev@zope.org',
      description='Cataloging and Indexing Framework for the Zope Toolkit',
      long_description=(
          read('README.rst')
          + '\n\n' +
          read('CHANGES.rst')
      ),
      keywords="zope3 catalog index",
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Web Environment',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Zope Public License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: PyPy',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Topic :: Internet :: WWW/HTTP',
          'Framework :: Zope :: 3',
      ],
      url='https://github.com/zopefoundation/zope.catalog',
      license='ZPL 2.1',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      namespace_packages=['zope'],
      extras_require={
          'test': tests_require,
          'docs': [
              'Sphinx',
              'repoze.sphinx.autointerface',
          ]
      },
      install_requires=[
          'setuptools',
          'persistent',
          'BTrees',
          'zope.annotation',
          'zope.intid',
          'zope.component >= 3.8',
          'zope.container',
          'zope.index>=3.5.0',
          'zope.interface',
          'zope.lifecycleevent',
          'zope.location',
          'zope.schema',
      ],
      tests_require=tests_require,
      test_suite='zope.catalog.tests.test_suite',
      include_package_data=True,
      zip_safe=False,
      )



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Catalog
"""
import BTrees
import zope.index.interfaces
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.container.btree import BTreeContainer
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.interface import implementer
from zope.intid.interfaces import IIntIdAddedEvent
from zope.intid.interfaces import IIntIdRemovedEvent
from zope.intid.interfaces import IIntIds
from zope.lifecycleevent import IObjectModifiedEvent
from zope.location import location
from zope.location.interfaces import ILocationInfo

from zope import component
from zope.catalog.interfaces import ICatalog
from zope.catalog.interfaces import ICatalogIndex
from zope.catalog.interfaces import INoAutoIndex
from zope.catalog.interfaces import INoAutoReindex


class ResultSet:
    """Lazily accessed set of objects."""

    def __init__(self, uids, uidutil):
        self.uids = uids
        self.uidutil = uidutil

    def __len__(self):
        return len(self.uids)

    def __iter__(self):
        for uid in self.uids:
            obj = self.uidutil.getObject(uid)
            yield obj


@implementer(ICatalog,
             IAttributeAnnotatable,
             zope.index.interfaces.IIndexSearch,
             )
class Catalog(BTreeContainer):

    family = BTrees.family32

    def __init__(self, family=None):
        super(Catalog, self).__init__()
        if family is not None:
            self.family = family

    def clear(self):
        for index in self.values():
            index.clear()

    def index_doc(self, docid, texts):
        """Register the data in indexes of this catalog."""
        for index in self.values():
            index.index_doc(docid, texts)

    def unindex_doc(self, docid):
        """Unregister the data from indexes of this catalog."""
        for index in self.values():
            index.unindex_doc(docid)

    def _visitSublocations(self):
        """Restricts the access to the objects that live within
        the nearest site if the catalog itself is locatable.
        """
        uidutil = None
        locatable = ILocationInfo(self, None)
        if locatable is not None:
            site = locatable.getNearestSite()
            sm = site.getSiteManager()
            uidutil = sm.queryUtility(IIntIds)
            if uidutil not in [c.component for c in sm.registeredUtilities()]:
                # we do not have a local inits utility
                uidutil = component.getUtility(IIntIds, context=self)
                for uid in uidutil:
                    obj = uidutil.getObject(uid)
                    if location.inside(obj, site):
                        yield uid, obj
                return
        if uidutil is None:
            uidutil = component.getUtility(IIntIds)
        for uid in uidutil:
            yield uid, uidutil.getObject(uid)

    def updateIndex(self, index):
        for uid, obj in self._visitSublocations():
            index.index_doc(uid, obj)

    def updateIndexes(self):
        for uid, obj in self._visitSublocations():
            for index in self.values():
                index.index_doc(uid, obj)

    def apply(self, query):
        results = []
        for index_name, index_query in query.items():
            index = self[index_name]
            r = index.apply(index_query)
            if r is None:
                continue
            if not r:
                # empty results
                return r
            results.append((len(r), r))

        if not results:
            # no applicable indexes, so catalog was not applicable
            return None

        results.sort(key=lambda x: x[0])  # order from smallest to largest

        _, result = results.pop(0)
        for _, r in results:
            _, result = self.family.IF.weightedIntersection(result, r)

        return result

    def searchResults(self, **searchterms):
        sort_index = searchterms.pop('_sort_index', None)
        limit = searchterms.pop('_limit', None)
        reverse = searchterms.pop('_reverse', False)
        results = self.apply(searchterms)
        if results is not None:
            if sort_index is not None:
                index = self[sort_index]
                if not zope.index.interfaces.IIndexSort.providedBy(index):
                    raise ValueError(
                        'Index %s does not support sorting.' %
                        sort_index)
                results = list(
                    index.sort(
                        results,
                        limit=limit,
                        reverse=reverse))
            else:
                if reverse or limit:
                    results = list(results)
                if reverse:
                    results.reverse()
                if limit:
                    del results[limit:]
            uidutil = component.getUtility(IIntIds)
            results = ResultSet(results, uidutil)
        return results


@component.adapter(ICatalogIndex, IObjectAddedEvent)
def indexAdded(index, event):
    """When an index is added to a catalog, we have to index existing objects

       When an index is added, we tell it's parent to index it:

         >>> class FauxCatalog:
         ...     updated = None
         ...     def updateIndex(self, index):
         ...         self.updated = index

         >>> class FauxIndex:
         ...     pass

         >>> index = FauxIndex()
         >>> index.__parent__ = FauxCatalog()

         >>> from zope.catalog.catalog import indexAdded
         >>> indexAdded(index, None)
         >>> index.__parent__.updated is index
         True
       """
    index.__parent__.updateIndex(index)


@component.adapter(IIntIdAddedEvent)
def indexDocSubscriber(event):
    """A subscriber to IntIdAddedEvent"""
    ob = event.object
    if INoAutoIndex.providedBy(ob):
        return
    for cat in component.getAllUtilitiesRegisteredFor(ICatalog, context=ob):
        id = component.getUtility(IIntIds, context=cat).getId(ob)
        cat.index_doc(id, ob)


@component.adapter(IObjectModifiedEvent)
def reindexDocSubscriber(event):
    """A subscriber to ObjectModifiedEvent"""
    ob = event.object
    if INoAutoReindex.providedBy(ob):
        return
    for cat in component.getAllUtilitiesRegisteredFor(ICatalog, context=ob):
        id = component.getUtility(IIntIds, context=cat).queryId(ob)
        if id is not None:
            cat.index_doc(id, ob)


@component.adapter(IIntIdRemovedEvent)
def unindexDocSubscriber(event):
    """A subscriber to IntIdRemovedEvent"""
    ob = event.object
    for cat in component.getAllUtilitiesRegisteredFor(ICatalog, context=ob):
        id = component.getUtility(IIntIds, context=cat).queryId(ob)
        if id is not None:
            cat.unindex_doc(id)



##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Catalog Interfaces
"""

import zope.container.constraints
import zope.container.interfaces
import zope.index.interfaces
import zope.interface
import zope.schema
from zope.i18nmessageid import ZopeMessageFactory as _


class ICatalogQuery(zope.interface.Interface):
    """Provides Catalog Queries."""

    def searchResults(**kw):
        """Search on the given indexes.

        Keyword arguments dictionary keys
        are index names and values are queries
        for these indexes.

        Keyword arguments has some special names,
        used by the catalog itself:

         * _sort_index - The name of index to sort
           results with. This index must implement
           zope.index.interfaces.IIndexSort.
         * _limit - Limit result set by this number,
           useful when used with sorting.
         * _reverse - Reverse result set, also
           useful with sorting.

        """


class ICatalogEdit(zope.index.interfaces.IInjection):
    """Allows one to manipulate the Catalog information."""

    def updateIndexes():
        """Reindex all objects."""


class ICatalogIndex(zope.index.interfaces.IInjection,
                    zope.index.interfaces.IIndexSearch,
                    ):
    """An index to be used in a catalog
    """

    __parent__ = zope.schema.Field()

    zope.container.constraints.containers('.ICatalog')


class ICatalog(ICatalogQuery, ICatalogEdit,
               zope.container.interfaces.IContainer):
    """Marker to describe a catalog in content space."""

    zope.container.constraints.contains(ICatalogIndex)


class IAttributeIndex(zope.interface.Interface):
    """I index objects by first adapting them to an interface, then
       retrieving a field on the adapted object.
    """

    interface = zope.schema.Choice(
        title=_(u"Interface"),
        description=_(u"Objects will be adapted to this interface"),
        vocabulary="Interfaces",
        required=False,
    )

    field_name = zope.schema.NativeStringLine(
        title=_(u"Field Name"),
        description=_(u"Name of the field to index"),
    )

    field_callable = zope.schema.Bool(
        title=_(u"Field Callable"),
        description=_(u"If true, then the field should be called to get the "
                      u"value to be indexed"),
    )


class INoAutoIndex(zope.interface.Interface):
    """Marker for objects that should not be automatically indexed"""


class INoAutoReindex(zope.interface.Interface):
    """Marker for objects that should not be automatically reindexed"""
